import java.io.*;
import java.util.*;

// ----------------------------------------------------------------------------

// All arrays (1-D and 2-D) have the same number of rows. Their
// indices start with 1, not 0.

class Beam {

    private String inputFile;

    private int numIntensityRows;  // r
    private int numIntensityCols;  // c
    private int numDifferenceCols; // c + 1

    private int[][] intensityMatrix, originalIntensityMatrix;
    private int[][] differenceMatrix, originalDifferenceMatrix;

    private int maxPointOfCurrentIntensityMatrix; 
    private int minPointOfCurrentIntensityMatrix;

    private int[][] intensityMatrixInProgress; // for debugging
    private boolean[] goodRow;

    private LinkedList[] rowShapeMatrixLists; // index i: all single-row shape 
                                              //          matrices for row i.
                                              // will be overwritten each 
                                              // iteration of removing a shape 
                                              // with a certain value 

    private LinkedList[] shapeMatrixLists;  // index i: all shape matrices with
                                            //          i rows 
                                            // will be overwritten each 
                                            // iteration of removing a shape 
                                            // with a certain value

    private final int heapSize                         = 10000;
    private final int numShapeMatricesToRemoveFromHeap = 1000;

    private FeasibleSoln bestSoln;       // store the best solution found so far

    // ------------------------------------------------------------------------

    public Beam(String inputFile1) {

	inputFile = inputFile1;

	intensityMatrix = Matrix.getIntensityMatrix(inputFile);
	originalIntensityMatrix = Matrix.getIntensityMatrix(inputFile);
	System.out.println("intensity matrix:");
	Matrix.printMatrix(intensityMatrix);

	differenceMatrix = Matrix.getDifferenceMatrix(intensityMatrix);
	originalDifferenceMatrix = Matrix.getDifferenceMatrix(originalIntensityMatrix);
	System.out.println("difference matrix:");
	Matrix.printMatrix(differenceMatrix);

	printSeparator();

	numIntensityRows = intensityMatrix.length - 1;
	numIntensityCols = intensityMatrix[0].length - 1;

	numDifferenceCols = numIntensityCols + 1;

	rowShapeMatrixLists = new LinkedList[numIntensityRows + 1];
	shapeMatrixLists    = new LinkedList[numIntensityRows + 1];

	for (int row = 1; row <= numIntensityRows; row++) {
	    rowShapeMatrixLists[row] = new LinkedList();
	    shapeMatrixLists[row]    = new LinkedList();
	}

	/******* FOR DEBUGGING ********/
	intensityMatrixInProgress = new int[numIntensityRows+1][numIntensityCols+1];
	for (int row = 1; row <= numIntensityRows; row++) {
	    for (int col = 1; col <= numIntensityCols; col++) {
		intensityMatrixInProgress[row][col] = intensityMatrix[row][col];
	    }
	}

	goodRow = new boolean[numIntensityRows + 1];

    }

    // ------------------------------------------------------------------------

    public static void main(String[] args) {

	if (args.length != 1)
	    fatalError("usage: java Beam inputFile");


	Beam beam = new Beam(args[0]);
	beam.optimize();
    }

    // ------------------------------------------------------------------------

    public void optimize() {
        FeasibleSoln parentSoln= new FeasibleSoln();
        FeasibleSoln promRegion= new FeasibleSoln(),				
                     compRegion= new FeasibleSoln();
        LinkedList neededShapes= new LinkedList();
        int neededShapeCnt=0;
        //Calendar cal;
		  
        // when shapeCountNP is odd: promising region, even: complementary region
        int maxShapes = 1;
        int shapeCountNP= 1;
        while( shapeCountNP<=2 || !( promRegion.getMemoryShapes().size()== maxShapes && 
                                     compRegion.getMemoryShapes().size()==maxShapes ) ){//shapeCountNP < maxShapes ){
            //long millis = cal.getTimeInMillis();
            if (shapeCountNP%2==1)
                System.out.println("Promising Region");
            else
                System.out.println("Complementary Region");
				
            int value;
            for (int i= 0; i < originalIntensityMatrix.length; i++)
                for (int j = 0; j < originalIntensityMatrix[i].length; j++){
                    intensityMatrix[i][j]= originalIntensityMatrix[i][j];
					 intensityMatrixInProgress[i][j]=  originalIntensityMatrix[i][j];
                }
            for (int i= 0; i < originalDifferenceMatrix.length; i++)
                for (int j = 0; j < originalDifferenceMatrix[i].length; j++)
                    differenceMatrix[i][j]= originalDifferenceMatrix[i][j];
            
            // force the set number of shapes- based on shapeCountNP for now
            System.out.println("printing the forced shapes");
            for (int i = 0; shapeCountNP%2==1 && i < promRegion.getMemoryShapes().size(); i++){
                System.out.println(promRegion.getMemoryShapes().get(i));
                removeShapeMatrixFromDifferenceMatrix((ShapeMatrix)promRegion.getMemoryShapes().get(i));
            }

            System.out.println("difference matrix:");
            Matrix.printMatrix(differenceMatrix);
				
            // disallow the set number of shapes- based on shapeCountNP for now
            ShapeMatrix disallowedShape = null;
            //for (int i = 0; i < 1000; i++)
            //    forcedShapesUsed[i]=false;
            while ((value= mostFrequentNonzeroAbsoluteDifference()) > 0) {
                int numForcedShapesUsed=0;
                ShapeMatrix s;
					 
                createShapeMatrices(value);
                // for the complementary region, if there is only one forcedShape not being used
                // then that shape cannot be allowed into the solution shapes
                if (shapeCountNP%2==0){
                    System.out.println("printing the disallowed shape");
                    disallowedShape= compRegion.getDisallowedShape();
                    System.out.println(disallowedShape);
                    compRegion.printMemoryShapes();
                    s = getShapeMatrixCandidateVer2(value, disallowedShape);
                } else {
                    System.out.println("There are no disallowed shapes for prom region");
                    s = getShapeMatrixCandidateVer2(value, null);
                }
                // if s is null, must force the remaining disallowed shape and start over
                if (s==null){ // should only happen when handling the complement solution
                    neededShapes.add(compRegion.getDisallowedShape());
                    neededShapeCnt++;
                    // setup for restart
                    for (int i= 0; i < originalIntensityMatrix.length; i++)
                        for (int j = 0; j < originalIntensityMatrix[i].length; j++){
                            intensityMatrix[i][j]= originalIntensityMatrix[i][j];
                            intensityMatrixInProgress[i][j]=  originalIntensityMatrix[i][j];
                        }
                    for (int i= 0; i < originalDifferenceMatrix.length; i++)
                        for (int j = 0; j < originalDifferenceMatrix[i].length; j++)
                            differenceMatrix[i][j]= originalDifferenceMatrix[i][j];
                            
                    FeasibleSoln tmp = new FeasibleSoln();
                    tmp.addMemoryShapes(compRegion.getMemoryShapes());
                    tmp.addMemoryShapeValues(compRegion.getMemoryShapeValues());
                    compRegion= new FeasibleSoln();
                    compRegion= tmp;
                    for (int i = 0; i < neededShapes.size(); i++){
                        System.out.println("Updating the disallowed shape as a needed shape");
                        removeShapeMatrixFromDifferenceMatrix((ShapeMatrix)(neededShapes.get(i)));
                        compRegion.updateShapeUsed((ShapeMatrix)(neededShapes.get(i)));
                        compRegion.addShape((ShapeMatrix)(neededShapes.get(i)),value);
                    }
                } else{
                    if (shapeCountNP%2==1)
                        promRegion.addShape(s,value);
                    else {
                        // see if s is a shape in the promising region (in forcedShapes
                        // if so: record it was used.
                        System.out.println("comparing the found shape with the forced shapes");
                        if (compRegion.canUseShape(s)){
                            System.out.println("Can Use the shape found");
                            compRegion.updateShapeUsed(s);
                            compRegion.addShape(s, value);
                        }
                    }
                    removeShapeMatrixFromDifferenceMatrix(s);
            
                    if (shapeCountNP%2==1)
                        System.out.println("shape matrix " + promRegion.getShapesPromRegion().size() + ":");
                    else
                        System.out.println("shape matrix " + compRegion.getShapesCompRegion().size() + ":");
                    System.out.println(s);
    
                    System.out.println("difference matrix:");
                    Matrix.printMatrix(differenceMatrix);
            
                    /***** FOR DEBUGGING ****/
                    System.out.println("new Intensity Matrix:");
                    Matrix.printMatrix(intensityMatrixInProgress);

                    printSeparator();                
                }
            }// end while
            
            //System.err.println("The current time in milliseconds is " + (cal.getTimeInMillis()-millis));
            if (shapeCountNP%2==1){
                System.err.println("Promising region:");
                System.out.println("beam time: " + (promRegion.getBOTmemory()+ promRegion.getBOTshapes()));
                System.err.print("total shape matrices: " + promRegion.getShapesPromRegionNoDups().size());
                maxShapes = promRegion.getShapesPromRegionNoDups().size();
                System.err.println("\t beam time: " + (promRegion.getBOTmemory()+ promRegion.getBOTshapes()) + "\n");
                // don't allow the comp region to use the forced shapes (including the last shape of the prom region
                compRegion.addMemoryShape((ShapeMatrix)(promRegion.getShape(0)), promRegion.getShapeValue(0));
                System.out.println("The shapes in memory of the comp region");
                compRegion.printMemoryShapes();
            } else {		
                System.err.println("Complementary region:");
                System.out.println("beam time: " + compRegion.getBOTshapes());
                System.err.print("total shape matrices: " + (compRegion.getShapesCompRegionNoDups().size()));
                System.err.println("\t beam time: " + compRegion.getBOTshapes() + "\n");
            }
            if (shapeCountNP%2==0){
                //complemenaryRegionBeamTime= beamTime;
                //complementaryRegionNumShapeMatrices= numShapeMatrices;
                // have information for both the promising region and the complementary region
                parentSoln= new FeasibleSoln();
                if ( (promRegion.getShapesPromRegionNoDups().size()*7 + (promRegion.getBOTmemory()+ promRegion.getBOTshapes()))
					  < (compRegion.getShapesCompRegionNoDups().size()*7 + compRegion.getBOTshapes()) ){
                    System.out.println("The promising region won");
                    System.err.println("The promising region won");
                    parentSoln= promRegion;
                    System.out.println("Print the winning promRegion");
                    parentSoln.printMemoryShapes();
                    parentSoln.printShapes();
                    System.out.println("Printing the shape adding");
                    System.out.println( (ShapeMatrix)(promRegion.getShape(0)));
                    parentSoln.addMemoryShape( (ShapeMatrix)(promRegion.getShape(0)),
                                                promRegion.getShapeValue(0));
                    parentSoln.removeFirstShape();
                    System.out.println("the parent shape(s): ");
                    parentSoln.printMemoryShapes();
                    parentSoln.printShapes();
                    // promising region won: keep next shape as the forced shape
                } else {
                    System.out.println("The complementary region won");
                    System.err.println("The complementary region won");
                    parentSoln= compRegion;
                    int numMemoryShapes= parentSoln.getMemoryShapes().size()+1;
                    System.out.println("The number of memoryShapes is " + numMemoryShapes);
                    // remove the unused MemoryShapes from the memorySpace
                    int remove = parentSoln.removeUnusedMemoryShapes(); // should only be one unused shape
                    System.out.println("Printing the used memory shapes");
                    parentSoln.printMemoryShapes();
                    // remove all shapes that were stored as memory shapes
                    for (int i = 0; i < parentSoln.getShapesCompRegion().size(); i++){
                        for (int j = 0; j < parentSoln.getMemoryShapes().size(); j++)
                            if ( ((ShapeMatrix)(parentSoln.getShapesCompRegion().get(i)))
                                .equal((ShapeMatrix)(parentSoln.getMemoryShapes().get(j))) ){
                                parentSoln.removeShape(i);
                                i--;
                                j= parentSoln.getMemoryShapes().size();
                            }
                        }				  
                    for (int i = 0; i < (numMemoryShapes-parentSoln.getMemoryShapes().size()); i++){
                        parentSoln.addMemoryShape((ShapeMatrix)(parentSoln.getShape(0)), parentSoln.getShapeValue(0));
                        parentSoln.removeFirstShape();
                    }
                    System.out.println("the parent shape(s): ");
                    parentSoln.printMemoryShapes();
                    parentSoln.printShapes();
                }// end else
                
                // set all memory shapes to unused
                parentSoln.reinitUsedAllowedShapes();

                promRegion= new FeasibleSoln();
                for (int i = 0; i < parentSoln.getMemoryShapes().size(); i++){
                    promRegion.addMemoryShape(parentSoln.getMemoryShape(i), parentSoln.getMemoryShapeValue(i));
                }
                compRegion= new FeasibleSoln();
                for (int i = 0; i < parentSoln.getMemoryShapes().size(); i++){
                    compRegion.addMemoryShape(parentSoln.getMemoryShape(i), parentSoln.getMemoryShapeValue(i));
                }
                System.out.println("At the end");
                compRegion.printMemoryShapes();
            }
            shapeCountNP++;
            
            System.err.print("pr shapes: " + (promRegion.getMemoryShapes().size()) + "   ");
            System.err.print("cr shapes: " + (compRegion.getMemoryShapes().size()) + "   ");
            System.err.println("maxShapes= " + maxShapes);

        }// end while
    }// end method

    // ------------------------------------------------------------------------

    void createShapeMatrices(int value) {
	for (int row = 1; row <= numIntensityRows; row++) {
	    // will get set to false in getRowShapeMatricesFromDifferenceRow() 
	    // if no entry in the difference row equals value
	    goodRow[row] = true; 
	}

	for (int row = 1; row <= numIntensityRows; row++) {
	    rowShapeMatrixLists[row]
		= getRowShapeMatricesFromDifferenceRow(row, value);
	}

	// --------------------------------------------------------------------

	Heap heap = new Heap(new ShapeMatrixComparator(), heapSize);

	for (int row = 1; row <= numIntensityRows; row++) {
	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);
	    while (li.hasNext()) {
		RowShapeMatrix r = (RowShapeMatrix) li.next();
		ShapeMatrix s
		    = new ShapeMatrix(numIntensityRows, numIntensityCols);

		s.value                    = value;
		s.price                    = r.price;
		s.start                    = row;  // initially each shape matrix
		s.end                      = row;  // will be a single row
		s.rowShapeMatrixArray[row] = r;
		heap.insert(s);
	    }
	}

	// get all single row shape matrices from the heap
	shapeMatrixLists[1]
	    = getShapeMatricesFromHeap(heap, numShapeMatricesToRemoveFromHeap, 1);


	// catch extremely rare case in which there were no decent rows
	if (shapeMatrixLists[1].isEmpty()) {
	    System.err.println("PATCHING LIST");
	    shapeMatrixLists[1] = patchGetShapeMatrices(value, heap);
	}


	int numDuplicates = 0;

	// --------------------------------------------------------------------

	// extend single row shapes to 2 rows
	// extend 2 row shapes to 3 rows ...
	// extend n-1 row shapes to n rows
	for (int s_row = 2; s_row <= numIntensityRows; s_row++) {
	    //System.err.println("There are " + shapeMatrixLists[s_row-1].size() + " shape matrices with " + (s_row-1) + " rows");

	    heap = new Heap(new ShapeMatrixComparator(), heapSize);

	    ListIterator li = shapeMatrixLists[s_row - 1].listIterator(0);
	    while (li.hasNext()) {
		ShapeMatrix s = (ShapeMatrix) li.next();

		// if shape is at top - extend downward 
		// NOTE: addRowsToShapeMatrix() will try to extend upward if it can't
		//       extend down
		if (s.end < numIntensityRows) {
		    addRowsToShapeMatrix(heap, s, true);
		} else if (s.start > 1) {
		    // shape is at bottom - extend upward
		    // only extend upward if there is no "good" row directly above - 
		    // this avoids repeating shapes
		    if (!goodRow[s.start-1]) {
			addRowsToShapeMatrix(heap, s, false);
		    } else {
			numDuplicates++;
			//li.remove();
		    }
		}
	    }

	    shapeMatrixLists[s_row] = getShapeMatricesFromHeap
		(heap, numShapeMatricesToRemoveFromHeap, s_row); 

	}

	//System.err.println(numDuplicates + " duplicates were removed");
    }

    // ------------------------------------------------------------------------

    // Returns at most the requested number of shape matrices from the
    // heap.
    private LinkedList getShapeMatricesFromHeap(Heap heap, int number, int numRows) {
	LinkedList list = new LinkedList();
	number          = (int) Math.min(number, heap.size());

	while (number-- > 0) {
	    ShapeMatrix s = (ShapeMatrix)heap.remove();
	    if (s.end-s.start >= numRows-1) list.add(s);
	}

	return list;
    }

    // ------------------------------------------------------------------------

    private LinkedList patchGetShapeMatrices (int value, Heap heap) {
	for (int row = 1; row <= numIntensityRows; row++) {
	    rowShapeMatrixLists[row]
		= getRowShapeMatricesFromDifferenceRowRelaxed(row, value);
	}

	// --------------------------------------------------------------------

	heap = new Heap(new ShapeMatrixComparator(), heapSize);

	for (int row = 1; row <= numIntensityRows; row++) {
	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);
	    while (li.hasNext()) {
		RowShapeMatrix r = (RowShapeMatrix) li.next();
		ShapeMatrix s
		    = new ShapeMatrix(numIntensityRows, numIntensityCols);

		s.value                    = value;
		s.price                    = r.price;
		s.start                    = row;  // initially each shape matrix
		s.end                      = row;  // will be a single row
		s.rowShapeMatrixArray[row] = r;
		heap.insert(s);
	    }
	}

	// get all single row shape matrices from the heap
	LinkedList list = getShapeMatricesFromHeap(heap, numShapeMatricesToRemoveFromHeap, 1);
	return list;

    }


    // ------------------------------------------------------------------------

    private LinkedList patchGetShapeMatricesOld (int value) {
	LinkedList list = new LinkedList();

	// If the value's absolute value appears in the difference
	// matrix row, create row shape matrices from the row.
	for (int row = 1; row <= numIntensityRows; row++) {
	    for (int colStart = 1; colStart <= numDifferenceCols-1; colStart++) {
		for (int colEnd = colStart+1; colEnd <= numDifferenceCols; colEnd++) {
		    if (differenceMatrix[row][colStart] == value || differenceMatrix[row][colEnd] == -value) {
			if (isDifferenceRowValid(row,value,colStart,colEnd)) {
			    int price = 1;
			    RowShapeMatrix r = new RowShapeMatrix(price, colStart, colEnd);
			    ShapeMatrix s = new ShapeMatrix(numIntensityRows, numIntensityCols);
			    
			    s.value                    = value;
			    s.price                    = r.price;
			    s.start                    = row;  // initially each shape matrix
			    s.end                      = row;  // will be a single row
			    s.rowShapeMatrixArray[row] = r;
			    list.add(s); 
			}
		    }
		}
	    }
	}

	if (list.size() == 0) {

	    System.err.print("\n\n\nstill no good rows!\n");
	    //System.err.print(Matrix.toString(intensityMatrix));
	    //System.err.print(Matrix.toString(intensityMatrixInProgress));
	    //System.err.print(Matrix.toString(differenceMatrix));
	    //System.err.print("\n\n\n");
	    System.exit(0);
	}

	return list;
    }


    // ------------------------------------------------------------------------

    private LinkedList getRowShapeMatricesFromDifferenceRow (int row, int value) {
	LinkedList list = new LinkedList();

	// If the value's absolute value appears in the difference
	// matrix row, create row shape matrices from the row.
	for (int col = 1; col <= numDifferenceCols; col++) {
	    if (Math.abs(differenceMatrix[row][col]) == value) {
		list.addAll(getRowShapeMatricesFromDifferenceRowSub
			    (row, col, value));
	    }
	}

	// If the value's absolute value did not appear in the
	// difference matrix row, do something else.
	if (list.size() == 0) {
	    goodRow[row] = false;
	    for (int i = 1; i <= numDifferenceCols; i++) {
		for (int j = i + 1; j <= numDifferenceCols; j++) {
		    if (differenceMatrix[row][i] != 0
			&& differenceMatrix[row][j] != 0
			&& isDifferenceRowValidVer2(row, value, i, j)) {

			list.add(new RowShapeMatrix(0, i, j));
		    }
		}
	    }
	}
	return list;
    }

    // ------------------------------------------------------------------------

    private LinkedList getRowShapeMatricesFromDifferenceRowRelaxed (int row, int value) {
	LinkedList list = new LinkedList();

	// If the value's absolute value appears in the difference
	// matrix row, create row shape matrices from the row.
	for (int col = 1; col <= numDifferenceCols; col++) {
	    if (Math.abs(differenceMatrix[row][col]) == value) {
		list.addAll(getRowShapeMatricesFromDifferenceRowSubRelaxed
			    (row, col, value));
	    }
	}

	// If the value's absolute value did not appear in the
	// difference matrix row, do something else.
	if (list.size() == 0) {
	    goodRow[row] = false;
	    for (int i = 1; i <= numDifferenceCols; i++) {
		for (int j = i + 1; j <= numDifferenceCols; j++) {
		    if (differenceMatrix[row][i] != 0
			&& differenceMatrix[row][j] != 0
			&& isDifferenceRowValidVer2(row, value, i, j)) {

			list.add(new RowShapeMatrix(0, i, j));
		    }
		}
	    }
	}
	return list;
    }

    // ------------------------------------------------------------------------

    private LinkedList getRowShapeMatricesFromDifferenceRowSub (int row, int col, int value) {
	LinkedList list = new LinkedList();

	if (differenceMatrix[row][col] > 0) { // Go right.
	    for (int j = col + 1; j <= numDifferenceCols; j++) {
		if (differenceMatrix[row][j] != 0
		    && isDifferenceRowValidVer2(row, value, col, j)) {

		    int price
			= 1 + ((differenceMatrix[row][j] == -value) ? 1 : 0);

		    list.add(new RowShapeMatrix(price, col, j));
		}
	    }
	} else { // Go left.
	    for (int j = 1; j < col; j++)
		if (differenceMatrix[row][j] != 0
		    && isDifferenceRowValidVer2(row, value, j, col)) {

		    int price
			= 1 + ((differenceMatrix[row][j] == value) ? 1 : 0);

		    // Do not include price = 2 values twice.
		    if (price < 2)
			list.add(new RowShapeMatrix(price, j, col));
		}
	}
	return list;
    }

    // ------------------------------------------------------------------------

    private LinkedList getRowShapeMatricesFromDifferenceRowSubRelaxed (int row, int col, int value) {
	LinkedList list = new LinkedList();

	if (differenceMatrix[row][col] > 0) { // Go right.
	    for (int j = col + 1; j <= numDifferenceCols; j++) {
		if (differenceMatrix[row][j] != 0
		    && isDifferenceRowValid(row, value, col, j)) {

		    int price
			= 1 + ((differenceMatrix[row][j] == -value) ? 1 : 0);

		    list.add(new RowShapeMatrix(price, col, j));
		}
	    }
	} else { // Go left.
	    for (int j = 1; j < col; j++)
		if (differenceMatrix[row][j] != 0
		    && isDifferenceRowValid(row, value, j, col)) {

		    int price
			= 1 + ((differenceMatrix[row][j] == value) ? 1 : 0);

		    // Do not include price = 2 values twice.
		    if (price < 2)
			list.add(new RowShapeMatrix(price, j, col));
		}
	}
	return list;
    }

    // ------------------------------------------------------------------------

    void addRowsToShapeMatrix(Heap heap, ShapeMatrix s, boolean top) {
	int begin1, end1, begin2, end2;
	ShapeMatrix new_s;

	// Add rows from the top of the shape matrix to the bottom.
	if (top) {
	    boolean tryExtendingUp = (s.start > 1); // only go up if there is a row above
	    
	    ListIterator li = rowShapeMatrixLists[s.end + 1].listIterator(0);
	    while (li.hasNext()) {
		RowShapeMatrix r = (RowShapeMatrix) li.next();

		begin1 = 1;
		end1   = numDifferenceCols;
		begin2 = r.left;
		end2   = r.right;

		if (s.rowShapeMatrixArray[s.end] != null) {
		    begin1 = s.rowShapeMatrixArray[s.end].left;
		    end1   = s.rowShapeMatrixArray[s.end].right;
		}

		// Check if the leaf constraint is satisfied. Check if
		// the end col of top leaf is at least the begin col
		// of the bottom leaf, and the end col of the bottom
		// leaf is at least the begin col of the top leaf.

		// --------b1--------e1-------- 2 rows of valid leaves
		// ------------b2------e2------

		if (end1 > begin2 && end2 > begin1) {
		    new_s = (ShapeMatrix) s.clone();
		    tryExtendingUp = false;
		    new_s.rowShapeMatrixArray[++new_s.end] = r;
		    new_s.price += r.price;

		    heap.insert(new_s);
		}
	    }


	    if (!tryExtendingUp) return;
	}


	// The rest of this method is for adding rows from the bottom
	// of the shape matrix to the top. Only does this if top == false
	ListIterator li = rowShapeMatrixLists[s.start - 1].listIterator(0);
	while (li.hasNext()) {
	    RowShapeMatrix r = (RowShapeMatrix) li.next();

	    begin1 = 1;
	    end1   = numDifferenceCols;
	    begin2 = r.left;
	    end2   = r.right;

	    if (s.rowShapeMatrixArray[s.start] != null) {
		begin1 = s.rowShapeMatrixArray[s.start].left;
		end1   = s.rowShapeMatrixArray[s.start].right;
	    }

	    // Check if the leaf constraint is satisfied.
   	    if (end1 > begin2 && end2 > begin1) {
		new_s = (ShapeMatrix) s.clone();
		new_s.rowShapeMatrixArray[--new_s.start] = r;
		new_s.price                             += r.price;

		heap.insert(new_s);
	    }
	}

	return;
    }

    // ------------------------------------------------------------------------

    // s is null if there is no more candidate shapes
    ShapeMatrix getShapeMatrixCandidateVer2(int value, ShapeMatrix disallowedShape) {
	// Let i be the index of the last list in the shape matrix
	// lists array that has a nonzero number of shape matrices
	// (that is, i is the max number of rows in a shape matrix we
	// can construct using the value). In the list, find the shape
	// matrix with the max price.

	int i;
	for (i = numIntensityRows; i >= 1; i--) {
	    if (shapeMatrixLists[i].size() > 0) break;
	}

	int top = topOfIntensityMatrix();
	int bottom = bottomOfIntensityMatrix();

	// System.err.println("\ntop: " + top + "   bottom: " + bottom);

	ShapeMatrix s = null;
	int sPrice = -1;
	int sAtTopOrBottom = 0;
      
	while (i >= 1) {
	    ListIterator li  = shapeMatrixLists[i].listIterator(0);
	    i--;
	    while (li.hasNext()) {
		ShapeMatrix temp = (ShapeMatrix) li.next();
		if (!straightEdge(temp)) continue;
                // disallow the remaining forced shape if have used all forced shapes but one
                // temp cannot equal disallowed shape  
                if (disallowedShape != null 
                    && temp.equal(disallowedShape)) {
						  System.out.println("This shape is disallowed");
						  continue;
					 }

		if (temp.price > sPrice) {
		    s = temp;
		    sPrice = s.price;
		    sAtTopOrBottom = shapeIsAtTopOfIntensityMatrix(s, top) + shapeIsAtBottomOfIntensityMatrix(s, bottom);
		} else if (temp.price == sPrice) {
		    int tempAtTopOrBottom = shapeIsAtTopOfIntensityMatrix(temp, top) + shapeIsAtBottomOfIntensityMatrix(temp, bottom);
		    
		    // check to see if temp is at top or bottom of remaining intensity matrix, when s wasn't
		    if (tempAtTopOrBottom > sAtTopOrBottom) {
			// System.err.println("temp was better");
			s = temp;
			sPrice = s.price;
			sAtTopOrBottom = tempAtTopOrBottom;
		    }
		}
	    }

	} 

	if (s != null) {	    
	    //System.err.println("top: " + s.start + "   bottom: " + s.end);
	} else {
	    System.err.println("no straight-edge found!");
            System.out.println("no straight-edge found!");
            s= null;
	    //System.err.println("There are " + shapeMatrixLists[1].size() + " single-row shape matrices in the heap");
	    //System.err.println("EXITING\n");
	    //System.exit(0);
	}
	return s;
    }


    boolean straightEdge(ShapeMatrix s) {
	boolean leftStraight = true;
	boolean rightStraight = true;
	int left, right;
	int middle = numIntensityRows/2;
	
	if(s.start <= middle && s.end >= middle+1) 
	    return true;

	if (s.start == s.end) 
	    return true;

	// otherwise must be in top half or bottom half - so check
	// for a straight edge at the left or right
	
	//check left edge
	RowShapeMatrix r = s.rowShapeMatrixArray[s.start];
	left = r.left;
	for(int i = s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.left != left) {
		leftStraight = false;
		break;
	    }
	}
	if(leftStraight) return true;
		
	// left edge wasn't straight
	// so check right edge
	r = s.rowShapeMatrixArray[s.start];
	right = r.right;
	for(int i = s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.right != right) {
		rightStraight = false;
		break;
	    }
	}
		
	return rightStraight;
    }


    // don't know what the s_row and s_col were useful for
    boolean straightEdgeAthula(ShapeMatrix s, int s_row, int s_col)
    {
	boolean leftStraight = true;
	boolean rightStraight = true;
	int left, right;
	int middle = numIntensityRows/2;
	int i;
	
	if(s.start+ s_row <= middle && s.end + s_row >= middle+1) 
	    return true;

	// otherwise must be in top half or bottom half - so check
	// for a straight edge at the left or right
	
	//check left edge
	RowShapeMatrix r = s.rowShapeMatrixArray[s.start];
	left = r.left;
	for(i=s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.left != left) {
		leftStraight = false;
		break;
	    }
	}
	if(leftStraight) return true;
		
	// left edge wasn't straight
	// so check right edge
	r = s.rowShapeMatrixArray[s.start];
	right = r.right;
	for(i=s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.right != right) {
		rightStraight = false;
		break;
	    }
	}
		
	return rightStraight;
    }


    /* DOESN'T IMPROVE SOLUTION */
    int topOfIntensityMatrix() {
	int i;
	for (i = 1; i <= numIntensityRows; i++) {
	    for (int j = 1; j <= numIntensityCols; j++) {
		if (differenceMatrix[i][j] != 0) {
		    return i;
		}
	    }
	}
	return i;
    }

    int bottomOfIntensityMatrix() {
	int i;
	for (i = numIntensityRows; i >= 1; i--) {
	    for (int j = 1; j <= numIntensityCols; j++) {
		if (differenceMatrix[i][j] != 0) {
		    return i;
		}
	    }
	}
	return i;

    }

    int shapeIsAtTopOfIntensityMatrix (ShapeMatrix s, int top) {
	if (s.start == top) return 1;
	return 0;
    }

    int shapeIsAtBottomOfIntensityMatrix (ShapeMatrix s, int bottom) {
	if (s.end == bottom) return 1;
	return 0;
    }


    /* DOESN'T IMPROVE SOLUTION */
    int maxHill (ShapeMatrix s) {
	int max = 0;
	int max_i = 0;
	int max_j = 0;

	for (int i = 1; i <= numIntensityRows; i++) {
	    int rowMax = 0;
	    int rowMax_j = 0;
	    for (int j = 1; j <= numIntensityCols; j++) {
		if (intensityMatrixInProgress[i][j] > rowMax) {
		    rowMax = intensityMatrixInProgress[i][j];
		    rowMax_j = j;
		}
	    }
		
	    if (rowMax > max) {
		max = rowMax;
		max_i = i;
		max_j = rowMax_j;
	    }
	}


	// if the shape matrix region contains the max intensity value return 1
	// otherwise return 0
	RowShapeMatrix r = s.rowShapeMatrixArray[max_i];
	if (r != null && 
	    r.left <= max_j &&
	    r.right >= max_j) {
	    return 1;
	}

	return 0;
    }


    /* NOT NECESSARY */
    int valueCount (ShapeMatrix s, int value) {
	int valCount = 0;

	// check left and right-most entries to see if they match value
	int row = s.start;
	while (row <= s.end) {
	    RowShapeMatrix r = s.rowShapeMatrixArray[row];
	    if (r != null) {
		if (differenceMatrix[row][r.left] == value) { 
		    valCount++;
		}
		if (differenceMatrix[row][r.right] == -value) { 
		    valCount++;
		}
	    }
	    row++;
	}

	return valCount;
    }


    // ------------------------------------------------------------------------

    boolean isDifferenceRowValid(int row, int value, int left, int right) {
	int sum = 0;

	for (int i = 1; i <= left; i++)
	    sum += differenceMatrix[row][i];

	if (sum < value)
	    return false;

	for (int i = left + 1; i < right; i++)
	    if ((sum += differenceMatrix[row][i]) < value)
		return false;


	for (int i = left; i < right; i++) {
	    if (intensityMatrixInProgress[row][i] < value) {
			  System.err.println("SHOULD HAVE BEEN FALSE!!!!!");
			  return false;
	    }
	}

	return true;
    }

    // ------------------------------------------------------------------------

    // avoid creating a "step" where none existed previously
    boolean isDifferenceRowValidVer2(int row, int value, int left, int right) {

	if (left >= right) return false;

	if (differenceMatrix[row][left] < value) return false;
	if (differenceMatrix[row][right] > -value) return false;

	int sum = value;

	for (int i = left+1; i < right; i++) {
	    sum+=differenceMatrix[row][i];
	    if (sum < value) return false;
	}

	for (int i = left; i < right; i++) {
	    if (intensityMatrixInProgress[row][i] < value) {
			  System.err.println("SHOULD HAVE BEEN FALSE!!!!!");
			  return false;
	    }
	}

	//System.err.println("found valid difference row");

	return true;
    }

    // ------------------------------------------------------------------------

    private void removeShapeMatrixFromDifferenceMatrix(ShapeMatrix s) {

	for (int row = s.start; row <= s.end; row++) {
	    RowShapeMatrix r = s.rowShapeMatrixArray[row];

	    if (r != null) {
		differenceMatrix[row][r.left ] -= s.value;
		differenceMatrix[row][r.right] += s.value;

		/******** FOR DEBUGGING **********/
		for (int col = r.left; col < r.right; col++) {
		    intensityMatrixInProgress[row][col] -= s.value;
		}
	    }
	}
    }

    // ------------------------------------------------------------------------

    private int mostFrequentNonzeroAbsoluteDifference() {

	// --------------------------------------------------------------------

	// Find the frequencies of the absolute values of the
	// difference matrix entries.

	HashMap hash = new HashMap();

	for (int row = 1; row <= numIntensityRows; row++) {
	    for (int col = 1; col <= numDifferenceCols; col++) {
		Integer absValue
		    = new Integer((int) Math.abs(differenceMatrix[row][col]));

		Integer frequency;

		if (!hash.containsKey(absValue))

		    frequency = new Integer(1);
		else
		    frequency = new Integer
			(1 + ((Integer) hash.get(absValue)).intValue());

		hash.put(absValue, frequency);
	    }
	}

	// --------------------------------------------------------------------

	// Sort the absolute values in decreasing order.

	int[] absValues   = new int[hash.size()];
	Iterator iterator = hash.keySet().iterator();

	for (int i = 0; i < absValues.length; i++) {
	    absValues[i] = ((Integer) iterator.next()).intValue();
	}

	Arrays.sort(absValues);

	// --------------------------------------------------------------------

	// Print the absolute values and their frequencies. Also, find
	// the absolute value with the max frequency and find the
	// number of nonzero absolute values.

	int maxFrequency         = -1;
	int mostFrequentAbsValue = -1;
	int totalNonzeroEntries  = 0;

	System.out.println("absVal\t" + "frequency");

	for (int i = 0; i < absValues.length; i++) {

	    int frequency
		= ((Integer) hash.get(new Integer(absValues[i]))).intValue();

	    if (absValues[i] > 0) {

		totalNonzeroEntries += frequency;

		if (frequency > maxFrequency) {

		    maxFrequency         = frequency;
		    mostFrequentAbsValue = absValues[i];
		}
	    }
	    System.out.println(absValues[i] + "\t" + frequency);
	}
	System.out.println();

	System.out.println("price\t" + totalNonzeroEntries
			   + " (total nonzero entries)\n");

	System.out.println("most frequent nonzero absolute value: "
			   + mostFrequentAbsValue);

	System.out.println();

	return mostFrequentAbsValue;
    }

    // ------------------------------------------------------------------------

    private void printRowShapeMatrixLists() {

	System.out.println("row shape matrix lists:");

	for (int row = 1; row <= numIntensityRows; row++) {

	    System.out.println("list " + row + ":");

	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);
	    int i = 0;

	    while (li.hasNext())

		System.out.println(i++ + "\t" + li.next());
	}
	System.out.println();
    }

    // ------------------------------------------------------------------------

    private void printShapeMatrixLists() {

	System.out.println("shape matrix lists:");

	for (int row = 1; row <= numIntensityRows; row++) {

	    System.out.println("list " + row + ":");

	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);
	    int i = 0;

	    while (li.hasNext())

		System.out.println(i++ + "\t" + li.next());
	}
	System.out.println();
    }

    // ------------------------------------------------------------------------

    private void printSeparator() {

	System.out.println("-------------------------------------------------------------------------------");
    }

    // ------------------------------------------------------------------------

    private static void fatalError(Object object) {

	System.out.println(object);
	System.exit(1);
    }
}
