import java.io.*;
import java.util.*;

// ----------------------------------------------------------------------------

// All arrays (1-D and 2-D) have the same number of rows. Their
// indices start with 1, not 0.

class BeamQuick {

    private String inputFile;

    private int numIntensityRows;  // r
    private int numIntensityCols;  // c
    private int numDifferenceCols; // c + 1

    private int[][] intensityMatrix;
    private int[][] differenceMatrix;

    private int maxPointOfCurrentIntensityMatrix; 
    private int minPointOfCurrentIntensityMatrix;

    private int[][] intensityMatrixInProgress; // for debugging
    private boolean[] goodRow;

    private LinkedList[] rowShapeMatrixLists; // index i: all single-row shape 
                                              //          matrices for row i.
                                              // will be overwritten each 
                                              // iteration of removing a shape 
                                              // with a certain value 

    private LinkedList[] shapeMatrixLists;  // index i: all shape matrices with
                                            //          i rows 
                                            // will be overwritten each 
                                            // iteration of removing a shape 
                                            // with a certain value

    private final int heapSize                         = 300;
    private final int numShapeMatricesToRemoveFromHeap = 100;

    // ------------------------------------------------------------------------

    public Beam(String inputFile1) {

	inputFile = inputFile1;

	intensityMatrix = Matrix.getIntensityMatrix(inputFile);
	//System.out.println("intensity matrix:");
	//Matrix.printMatrix(intensityMatrix);

	differenceMatrix = Matrix.getDifferenceMatrix(intensityMatrix);
	//System.out.println("difference matrix:");
	//Matrix.printMatrix(differenceMatrix);

	//printSeparator();

	numIntensityRows = intensityMatrix.length - 1;
	numIntensityCols = intensityMatrix[0].length - 1;

	numDifferenceCols = numIntensityCols + 1;

	rowShapeMatrixLists = new LinkedList[numIntensityRows + 1];
	shapeMatrixLists    = new LinkedList[numIntensityRows + 1];

	for (int row = 1; row <= numIntensityRows; row++) {
	    rowShapeMatrixLists[row] = new LinkedList();
	    shapeMatrixLists[row]    = new LinkedList();
	}

	/******* FOR DEBUGGING ********/
	intensityMatrixInProgress = new int[numIntensityRows+1][numIntensityCols+1];
	for (int row = 1; row <= numIntensityRows; row++) {
	    for (int col = 1; col <= numIntensityCols; col++) {
		intensityMatrixInProgress[row][col] = intensityMatrix[row][col];
	    }
	}

	goodRow = new boolean[numIntensityRows + 1];

    }

    // ------------------------------------------------------------------------

    public static void main(String[] args) {

	if (args.length != 1)
	    fatalError("usage: java Beam inputFile");


	BeamQuick beam = new BeamQuick(args[0]);
	beam.optimize();
    }

    // ------------------------------------------------------------------------

    public void optimize() {

	int value;
	int numShapeMatrices = 0;
	int beamTime         = 0;

	while ((value = mostFrequentNonzeroAbsoluteDifference()) > 0) {

	    numShapeMatrices++;
	    beamTime += value;

	    createShapeMatrices(value);

	    ShapeMatrix s = getShapeMatrixCandidateVer2(value);

	    removeShapeMatrixFromDifferenceMatrix(s);

	    //System.out.println("shape matrix " + numShapeMatrices + ":");
	    //System.out.println(s);

	    //System.out.println("difference matrix:");
	    //Matrix.printMatrix(differenceMatrix);

	    /***** FOR DEBUGGING ****/
	    //System.out.println("new Intensity Matrix:");
	    //Matrix.printMatrix(intensityMatrixInProgress);

	    printSeparator();
	}

	//System.out.println("beam time: " + beamTime);

	//System.err.print("total shape matrices: " + numShapeMatrices);
	//System.err.println("\t beam time: " + beamTime + "\n");
    }

    // ------------------------------------------------------------------------

    void createShapeMatrices(int value) {
	//System.err.println("createShapeMatrices");

	for (int row = 1; row <= numIntensityRows; row++) {
	    // will get set to false in getRowShapeMatricesFromDifferenceRow() 
	    // if no entry in the difference row equals value
	    goodRow[row] = true; 
	}

	for (int row = 1; row <= numIntensityRows; row++) {
	    rowShapeMatrixLists[row]
		= getRowShapeMatricesFromDifferenceRow(row, value);
	}

	// --------------------------------------------------------------------

	Heap heap = new Heap(new ShapeMatrixComparator(), heapSize);

	for (int row = 1; row <= numIntensityRows; row++) {
	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);

	    while (li.hasNext()) {

		RowShapeMatrix r = (RowShapeMatrix) li.next();

		ShapeMatrix s
		    = new ShapeMatrix(numIntensityRows, numIntensityCols);

		s.value                    = value;
		s.price                    = r.price;
		s.start                    = row;  // initially each shape matrix
		s.end                      = row;  // will be a single row
		s.rowShapeMatrixArray[row] = r;
		heap.insert(s);
	    }
	}

	// get all single row shape matrices from the heap
	shapeMatrixLists[1]
	    = getShapeMatricesFromHeap(heap, numShapeMatricesToRemoveFromHeap, 1);

	// System.err.println("There are " + shapeMatrixLists[1].size() + " single-row shape matrices in the heap");

	int numDuplicates = 0;

	// --------------------------------------------------------------------

	// extend single row shapes to 2 rows
	// extend 2 row shapes to 3 rows ...
	// extend n-1 row shapes to n rows
	for (int s_row = 2; s_row <= numIntensityRows; s_row++) {
	    //System.err.println("There are " + shapeMatrixLists[s_row-1].size() + " shape matrices with " + (s_row-1) + " rows");

	    heap = new Heap(new ShapeMatrixComparator(), heapSize);

	    ListIterator li = shapeMatrixLists[s_row - 1].listIterator(0);
	    while (li.hasNext()) {
		ShapeMatrix s = (ShapeMatrix) li.next();

		// if shape is at top - extend downward
		if (s.end < numIntensityRows) {
		    addRowsToShapeMatrix(heap, s, true);
		} else if (s.start > 1) {
		    // shape is at bottom - extend upward
		    // only extend upward if there is no "good" row directly above - 
		    // this avoids repeating shapes
		    if (!goodRow[s.start-1]) {
			addRowsToShapeMatrix(heap, s, false);
		    } else {
			numDuplicates++;
			li.remove();
		    }

		    //addRowsToShapeMatrix(heap, s, false);
		}
	    }

	    shapeMatrixLists[s_row] = getShapeMatricesFromHeap
		(heap, numShapeMatricesToRemoveFromHeap, s_row); 

	}

	//System.err.println(numDuplicates + " duplicates were removed");
    }

    // ------------------------------------------------------------------------

    // Returns at most the requested number of shape matrices from the
    // heap.

    private LinkedList getShapeMatricesFromHeap(Heap heap, int number, int numRows) {

	LinkedList list = new LinkedList();
	number          = (int) Math.min(number, heap.size());

	while (number-- > 0) {
	    ShapeMatrix s = (ShapeMatrix)heap.remove();
	    if (s.end-s.start >= numRows-1) list.add(s);
	}

	return list;
    }

    // ------------------------------------------------------------------------

    private LinkedList getRowShapeMatricesFromDifferenceRow (int row, int value) {
	//System.err.println("getRowShapeMatricesFromDifferenceRow");

	LinkedList list = new LinkedList();

	// If the value's absolute value appears in the difference
	// matrix row, create row shape matrices from the row.
	for (int col = 1; col <= numDifferenceCols; col++) {
	    if (Math.abs(differenceMatrix[row][col]) == value) {
		list.addAll(getRowShapeMatricesFromDifferenceRowSub
			    (row, col, value));
	    }
	}

	// If the value's absolute value did not appear in the
	// difference matrix row, do something else.
	if (list.size() == 0) {
	    goodRow[row] = false;
	    for (int i = 1; i <= numDifferenceCols; i++) {
		for (int j = i + 1; j <= numDifferenceCols; j++) {
		    if (differenceMatrix[row][i] != 0
			&& differenceMatrix[row][j] != 0
			&& isDifferenceRowValidVer2(row, value, i, j)) {

			list.add(new RowShapeMatrix(0, i, j));
		    }
		}
	    }
	}
	return list;
    }

    // ------------------------------------------------------------------------

    private LinkedList getRowShapeMatricesFromDifferenceRowSub (int row, int col, int value) {
	//System.err.println("getRowShapeMatricesFromDifferenceRowSub");
	
	LinkedList list = new LinkedList();

	if (differenceMatrix[row][col] > 0) { // Go right.
	    for (int j = col + 1; j <= numDifferenceCols; j++) {
		if (differenceMatrix[row][j] != 0
		    && isDifferenceRowValidVer2(row, value, col, j)) {

		    int price
			= 1 + ((differenceMatrix[row][j] == -value) ? 1 : 0);

		    list.add(new RowShapeMatrix(price, col, j));
		}
	    }
	} else { // Go left.
	    for (int j = 1; j < col; j++)
		if (differenceMatrix[row][j] != 0
		    && isDifferenceRowValidVer2(row, value, j, col)) {

		    int price
			= 1 + ((differenceMatrix[row][j] == value) ? 1 : 0);

		    // Do not include price = 2 values twice.
		    if (price < 2)
			list.add(new RowShapeMatrix(price, j, col));
		}
	}
	return list;
    }

    // ------------------------------------------------------------------------

    void addRowsToShapeMatrix(Heap heap, ShapeMatrix s, boolean top) {

	//System.err.println("addRowsToShapeMatrix");

	int begin1, end1, begin2, end2;
	boolean isLeafConstraintUnsatisfiable = true;
	ShapeMatrix new_s;

	// Add rows from the top of the shape matrix to the bottom.
	if (top) {
	    boolean tryExtendingUp = false;
	    
	    ListIterator li = rowShapeMatrixLists[s.end + 1].listIterator(0);
	    while (li.hasNext()) {
		RowShapeMatrix r = (RowShapeMatrix) li.next();

		begin1 = 1;
		end1   = numDifferenceCols;
		begin2 = r.left;
		end2   = r.right;

		if (s.rowShapeMatrixArray[s.end] != null) {
		    begin1 = s.rowShapeMatrixArray[s.end].left;
		    end1   = s.rowShapeMatrixArray[s.end].right;
		}

		// Check if the leaf constraint is satisfied. Check if
		// the end col of top leaf is at least the begin col
		// of the bottom leaf, and the end col of the bottom
		// leaf is at least the begin col of the top leaf.

		// --------b1--------e1-------- 2 rows of valid leaves
		// ------------b2------e2------

		if (end1 > begin2 && end2 > begin1) {
		    new_s = (ShapeMatrix) s.clone();

		    isLeafConstraintUnsatisfiable          = false;
		    new_s.rowShapeMatrixArray[++new_s.end] = r;
		    new_s.price                           += r.price;

		    heap.insert(new_s);
		}
	    }

	    if (isLeafConstraintUnsatisfiable) {
		//new_s = (ShapeMatrix) s.clone();
		//new_s.rowShapeMatrixArray[++new_s.end] = null;
		//heap.insert(new_s);
		if (s.start > 1) tryExtendingUp = true;
	    }
	    if (!tryExtendingUp) return;
	}


	// The rest of this method is for adding rows from the bottom
	// of the shape matrix to the top. Only does this if top == false
	ListIterator li = rowShapeMatrixLists[s.start - 1].listIterator(0);
	while (li.hasNext()) {
	    RowShapeMatrix r = (RowShapeMatrix) li.next();

	    begin1 = 1;
	    end1   = numDifferenceCols;
	    begin2 = r.left;
	    end2   = r.right;

	    if (s.rowShapeMatrixArray[s.start] != null) {
		begin1 = s.rowShapeMatrixArray[s.start].left;
		end1   = s.rowShapeMatrixArray[s.start].right;
	    }

	    // Check if the leaf constraint is satisfied.
   	    if (end1 > begin2 && end2 > begin1) {
		new_s = (ShapeMatrix) s.clone();

		isLeafConstraintUnsatisfiable            = false;
		new_s.rowShapeMatrixArray[--new_s.start] = r;
		new_s.price                             += r.price;

		heap.insert(new_s);
	    }
	}

	if (isLeafConstraintUnsatisfiable) {
	    // can't be extended further
	    /*
	    new_s = (ShapeMatrix) s.clone();
	    new_s.rowShapeMatrixArray[--new_s.start] = null;
	    heap.insert(new_s);
	    */
	}
	return;
    }

    // ------------------------------------------------------------------------

    ShapeMatrix getShapeMatrixCandidate() {

	// Let i be the index of the last list in the shape matrix
	// lists array that has a nonzero number of shape matrices
	// (that is, i is the max number of rows in a shape matrix we
	// can construct using the value). In the list, find the shape
	// matrix with the max price.

	int i;
	for (i = numIntensityRows; i >= 1; i--)

	    if (shapeMatrixLists[i].size() > 0)
		break;

	ShapeMatrix s   = (ShapeMatrix) shapeMatrixLists[i].getFirst();
	ListIterator li = shapeMatrixLists[i].listIterator(0);

	while (li.hasNext()) {

	    ShapeMatrix temp = (ShapeMatrix) li.next();

	    if (temp.price > s.price)
		s = temp;
	}
	return s;
    }


    // ------------------------------------------------------------------------

    ShapeMatrix getShapeMatrixCandidateVer2(int value) {

	//System.err.println("getShapeMatrixCandidateVer2");

	// Let i be the index of the last list in the shape matrix
	// lists array that has a nonzero number of shape matrices
	// (that is, i is the max number of rows in a shape matrix we
	// can construct using the value). In the list, find the shape
	// matrix with the max price.

	int i;
	for (i = numIntensityRows; i >= 1; i--) {
	    if (shapeMatrixLists[i].size() > 0) break;
	}

	int top = topOfIntensityMatrix();
	int bottom = bottomOfIntensityMatrix();

	// System.err.println("\ntop: " + top + "   bottom: " + bottom);

	ShapeMatrix s = null;
	int sPrice = -1;
      
	while (i >= 1) {
	    ListIterator li  = shapeMatrixLists[i].listIterator(0);
	    i--;
	    while (li.hasNext()) {
		ShapeMatrix temp = (ShapeMatrix) li.next();
		if (!straightEdge(temp)) continue;

		if (temp.price > sPrice) {
		    s = temp;
		    sPrice = s.price;
		}
	    }

	} 

	if (s != null) {	    
	    //System.err.println("top: " + s.start + "   bottom: " + s.end);
	} else {
	    System.err.println("ERROR: no straight-edge found!");
	    System.err.println("There are " + shapeMatrixLists[1].size() + " single-row shape matrices in the heap");
	    System.err.println("EXITING\n");
	    System.exit(0);
	}
	return s;
    }


    boolean straightEdge(ShapeMatrix s) {
	boolean leftStraight = true;
	boolean rightStraight = true;
	int left, right;
	int middle = numIntensityRows/2;
	
	if(s.start <= middle && s.end >= middle+1) 
	    return true;

	if (s.start == s.end) 
	    return true;

	// otherwise must be in top half or bottom half - so check
	// for a straight edge at the left or right
	
	//check left edge
	RowShapeMatrix r = s.rowShapeMatrixArray[s.start];
	left = r.left;
	for(int i = s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.left != left) {
		leftStraight = false;
		break;
	    }
	}
	if(leftStraight) return true;
		
	// left edge wasn't straight
	// so check right edge
	r = s.rowShapeMatrixArray[s.start];
	right = r.right;
	for(int i = s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.right != right) {
		rightStraight = false;
		break;
	    }
	}
		
	return rightStraight;
    }


    // don't know what the s_row and s_col were useful for
    boolean straightEdgeAthula(ShapeMatrix s, int s_row, int s_col)
    {
	boolean leftStraight = true;
	boolean rightStraight = true;
	int left, right;
	int middle = numIntensityRows/2;
	int i;
	
	if(s.start+ s_row <= middle && s.end + s_row >= middle+1) 
	    return true;

	// otherwise must be in top half or bottom half - so check
	// for a straight edge at the left or right
	
	//check left edge
	RowShapeMatrix r = s.rowShapeMatrixArray[s.start];
	left = r.left;
	for(i=s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.left != left) {
		leftStraight = false;
		break;
	    }
	}
	if(leftStraight) return true;
		
	// left edge wasn't straight
	// so check right edge
	r = s.rowShapeMatrixArray[s.start];
	right = r.right;
	for(i=s.start+1; i <= s.end; i++) {
	    r = s.rowShapeMatrixArray[i];
	    if(r.right != right) {
		rightStraight = false;
		break;
	    }
	}
		
	return rightStraight;
    }


    // ------------------------------------------------------------------------

    boolean isDifferenceRowValid(int row, int value, int left, int right) {
	int sum = 0;

	for (int i = 1; i <= left; i++)
	    sum += differenceMatrix[row][i];

	if (sum < value)
	    return false;

	for (int i = left + 1; i < right; i++)

	    if ((sum += differenceMatrix[row][i]) < value)
		return false;

	return true;
    }

    // ------------------------------------------------------------------------

    // avoid creating a "step" where none existed previously
    boolean isDifferenceRowValidVer2(int row, int value, int left, int right) {

	if (left >= right) return false;

	if (differenceMatrix[row][left] < value) return false;
	if (differenceMatrix[row][right] > -value) return false;

	int sum = differenceMatrix[row][left];

	for (int i = left+1; i < right; i++) {
	    if ((sum+=differenceMatrix[row][i]) < value) return false;
	}

	//System.err.println("found valid difference row");

	return true;
    }

    // ------------------------------------------------------------------------

    private void removeShapeMatrixFromDifferenceMatrix(ShapeMatrix s) {

	for (int row = s.start; row <= s.end; row++) {

	    RowShapeMatrix r = s.rowShapeMatrixArray[row];

	    if (r != null) {

		differenceMatrix[row][r.left ] -= s.value;
		differenceMatrix[row][r.right] += s.value;

		/******** FOR DEBUGGING **********/
		for (int col = r.left; col < r.right; col++) {
		    intensityMatrixInProgress[row][col] -= s.value;
		}

	    }
	}
    }

    // ------------------------------------------------------------------------

    private int mostFrequentNonzeroAbsoluteDifference() {

	// --------------------------------------------------------------------

	// Find the frequencies of the absolute values of the
	// difference matrix entries.

	HashMap hash = new HashMap();

	for (int row = 1; row <= numIntensityRows; row++) {
	    for (int col = 1; col <= numDifferenceCols; col++) {
		Integer absValue
		    = new Integer((int) Math.abs(differenceMatrix[row][col]));

		Integer frequency;

		if (!hash.containsKey(absValue))

		    frequency = new Integer(1);
		else
		    frequency = new Integer
			(1 + ((Integer) hash.get(absValue)).intValue());

		hash.put(absValue, frequency);
	    }
	}

	// --------------------------------------------------------------------

	// Sort the absolute values in decreasing order.

	int[] absValues   = new int[hash.size()];
	Iterator iterator = hash.keySet().iterator();

	for (int i = 0; i < absValues.length; i++) {
	    absValues[i] = ((Integer) iterator.next()).intValue();
	}

	Arrays.sort(absValues);

	// --------------------------------------------------------------------

	// Print the absolute values and their frequencies. Also, find
	// the absolute value with the max frequency and find the
	// number of nonzero absolute values.

	int maxFrequency         = -1;
	int mostFrequentAbsValue = -1;
	int totalNonzeroEntries  = 0;

	System.out.println("absVal\t" + "frequency");

	for (int i = 0; i < absValues.length; i++) {

	    int frequency
		= ((Integer) hash.get(new Integer(absValues[i]))).intValue();

	    if (absValues[i] > 0) {

		totalNonzeroEntries += frequency;

		if (frequency > maxFrequency) {

		    maxFrequency         = frequency;
		    mostFrequentAbsValue = absValues[i];
		}
	    }
	    System.out.println(absValues[i] + "\t" + frequency);
	}
	System.out.println();

	System.out.println("price\t" + totalNonzeroEntries
			   + " (total nonzero entries)\n");

	System.out.println("most frequent nonzero absolute value: "
			   + mostFrequentAbsValue);

	System.out.println();

	return mostFrequentAbsValue;
    }

    // ------------------------------------------------------------------------

    private void printRowShapeMatrixLists() {

	System.out.println("row shape matrix lists:");

	for (int row = 1; row <= numIntensityRows; row++) {

	    System.out.println("list " + row + ":");

	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);
	    int i = 0;

	    while (li.hasNext())

		System.out.println(i++ + "\t" + li.next());
	}
	System.out.println();
    }

    // ------------------------------------------------------------------------

    private void printShapeMatrixLists() {

	System.out.println("shape matrix lists:");

	for (int row = 1; row <= numIntensityRows; row++) {

	    System.out.println("list " + row + ":");

	    ListIterator li = rowShapeMatrixLists[row].listIterator(0);
	    int i = 0;

	    while (li.hasNext())

		System.out.println(i++ + "\t" + li.next());
	}
	System.out.println();
    }

    // ------------------------------------------------------------------------

    private void printSeparator() {

	System.out.println("-------------------------------------------------------------------------------");
    }

    // ------------------------------------------------------------------------

    private static void fatalError(Object object) {

	System.out.println(object);
	System.exit(1);
    }
}
